# Pattern Breeder, a generalisation of Ed Fredkin's pattern replicating XOR rule
# Breed the seed pattern in layer 1 with the neighbourhood defined by the pattern in layer 2.
# Author: Tony Smith (ts@meme.com.au), January 2011.

# Emulating Pattern Breeder as released at Macworld Expo January 1986
# and described in Scientific American that September.

# Updated February 2011 to add log n speed generation of an entered target iteration number
# and core Golly-standard keyboard shortcut interaction.

# Note: While there could be performance improvement through invoking rule trees for 3x3
# sub-neighbourhoods containing more than one counted location, identifying such is
# non-trivial, especially in edge cases, as well as the need to preinstall 502 rule trees.

use strict;
use Time::HiRes qw ( sleep );

g_exit("Needs a seed in layer 1 and a neighbourhood in layer 2!") unless g_numlayers() == 2;
g_setlayer(1);
my $neighbourhood = g_getcells( g_getrect() );
g_exit("Needs a neighbourhood defining pattern in layer 2!") unless scalar @$neighbourhood;
my @dxy;
if ( scalar @$neighbourhood % 2 ) { # multistate
  while ( my ( $x, $y, $state ) = splice( @$neighbourhood, 0, 3 ) ) {
    if ( $state ) { push ( @dxy, $x, $y ) }
  }
}
else {
  while ( my ( $x, $y ) = splice( @$neighbourhood, 0, 2 ) ) { push ( @dxy, $x, $y ) }
}
g_setlayer(0);
my @bounds = g_getrect();
g_exit("Needs a seed pattern in layer 1!") unless scalar @bounds;

my $params = g_getstring("Enter target iteration number to jump to\n".
                         "and optional seconds to delay starting:",
                         "0 2.5", "Jump to iteration");
my ( $start_at, $sleep_for ) = split( /\s/, $params );
( $start_at, $sleep_for ) = ( 0, 2.5 ) unless defined $start_at;
g_exit("Target ($start_at) must be a whole number.") unless $start_at == int($start_at);
my $todo = $start_at;
if ( $todo ) {
  my $bit = 1;
  while ( $todo ) {
    my $doing = $todo & $bit;
    if ( $doing ) {
      &iterate( $bit );
      $todo -= $bit;
    }
    $bit *= 2;
  }
  g_setgen( $start_at );
  g_fit() unless g_visrect( @bounds );
  g_update();
}
my $status;
if ( defined $sleep_for ) {
  sleep( $sleep_for );
  &status( 'breeding' );
}
else {
  &status( 'paused' );
}

my $oldsecs = time;
while ( 1 ) {
  my $key =  g_getkey(); # allow keyboard interaction
  if ( $status eq 'breeding' ) {
    if ( $key eq chr(13) ) { # return
      &status( 'paused' );
    }
    else {
      my $step = &iterate();
      g_setgen( "+$step" );
      &status( 'breeding' );
    }
  }
  elsif ( $key eq chr(13) ) { # return
    my $step = &iterate();
    g_setgen( "+$step" );
    &status( 'breeding' );
  }
  elsif ( $key eq ' ' ) { # space
    &iterate( 1 );
    g_setgen( "+1" );
    &status( 'paused' );
  }
  elsif ( $key eq chr(9) ) { # tab
    my $step = &iterate();
    g_setgen( "+$step" );
    &status( 'paused' );
  }
  elsif ( $key eq '+' ) { # faster
    g_setstep( g_getstep() + 1 );
  }
  elsif ( $key eq '-' ) { # slower
    g_setstep( g_getstep() - 1 );
  }
  else { g_dokey( $key ) }
  @bounds = g_getrect();
  my $newsecs = time;
  if ( $newsecs - $oldsecs >= 1.0 ) { # show update every second, allowing view changes while paused
    g_fit() unless g_visrect( @bounds ) or $status eq 'paused' ;
    g_update();
    $oldsecs = $newsecs;
  }
}
g_fit() unless g_visrect( @bounds );

sub iterate { # relies on global @dxy, @bounds, 
  my $bit = shift( @_ );
  $bit = g_getbase() ** g_getstep() unless defined $bit;
  g_select( @bounds );
  g_cut;
  g_select();
  my ( $x0, $y0 ) = @bounds;
  my @todo = @dxy;
  while ( my ( $dx, $dy ) = splice( @todo, 0, 2 ) ) {
    g_paste( $x0 + $dx * $bit, $y0 + $dy * $bit, "xor" );
  }
  @bounds = g_getrect();
  return $bit;
}

sub status { # relies on global $status
  $status = shift( @_ );
  if ( $status eq 'breeding' ) {
    g_show("Breeding... (hit return to pause or escape to abort)");
  }
  elsif ( $status eq  'paused' ) {
    g_show("Paused... (hit return to start breeding, space to breed next, tab to breed step or escape to abort)");
  }
  else {
    g_show("Done");
  }
}